#!/usr/local/ActivePerl-5.28/bin/perl
#!/usr/bin/perl

# Hacky first attempt at animating X-machine trace log using curses.

use strict;
use warnings;
use Curses;
use Scalar::Util;
use Time::HiRes;

# Usage:
#   perl trace.pl prog.xmout prog.log
#

my $debug = 0;

my ($xmout_file, $log_file) = @ARGV;

defined $xmout_file && defined $log_file or
    die "Usage: trace.pl prog.xmout prog.log";

open( MEM, "<", $xmout_file ) or
    die "Cannot open object file '$xmout_file', $!";

open( TRACE, "<", $log_file ) or
    die "Cannot open trace log file '$log_file', $!";

my $debug_file = "debug.out";
$debug or $debug_file = "/dev/null";

open( DEBUG, ">", $debug_file ) or
    die "Cannot open debug file '$debug_file', $!";

#
#   Fetch the initial contents of memory from .xmout file
#
my $mem = [ ];

# Get starting addr

my $addr = <MEM>;
defined $addr or
    die "Undefined addr header in object file '$xmout_file'";
chomp($addr);
$addr >= 0 or
    die "Invalid addr header '$addr' in object file '$xmout_file'";

# Get starting cell count

my $ncells = <MEM>;
defined $ncells or
    die "Undefined cell count header in object file '$xmout_file'";
chomp($ncells);
$ncells >= 0 or
    die "Invalid cell count header '$ncells' in object file '$xmout_file'";

# Load memory

for (my $i=0; $i < $ncells; $i++ ) {
    my $line = <MEM>;
    last unless $line;

    chomp($line);
    $debug and printf "%s\n", $line;

    $mem->[$addr] = $line; 
    $addr++;
    }

# This is only an initial estimate, memory can grow beyond this, and 
# so you may lose row address labels.  We could detect this and add the
# missing labels.

my $mem_max = @$mem;

#
# Fetch the execution trace from the .log file
#
# Fields are: [ $idx, $inst_num, $pc, $d, $a, $v ];
#   idx - sequence number of memory operation, first op starts at 0
#   inst_num - sequence number of instruction being executed, starts at 0
#   pc - value of program counter at time of mem operation
#   d - type of operation R-read, W-write
#   a - address of memory cell
#   v - value of memory cell after operation complete. I.e. is
#       the value read or written.

my $trace = [ ];

while ( my $line = <TRACE> ) {
    chomp($line);
    next if $line =~ /IDX/;
    next if $line =~ /^\s*$/;

    $line =~ s/^\s+//;
    $line =~ s/\s+$//;


    my ($idx, $inst_num, $pc, $d, $a, $v) = split(/\s+/, $line);

    $debug and 
        printf "%s %s %s %s %d %s\n", $idx, $inst_num, $pc, $d, $a, $v;

    push @$trace, [ $idx, $inst_num, $pc, $d, $a, $v ];
    }

my $trace_step_max = @$trace;

#
#   Set up the curses screen
#

# Layout the memory portion of known addresses in $nrows of $ncols
# per row.  Using 10 per row, we can use easily readable decimal addresses
# memory: 0 ... $mem_max-1

my $ncols = 10;
my $nrows = 1 + int( $mem_max / $ncols );

#
#   A hacked ADT, not an object yet.  So a screen is just a hash
#   containing all the stuff needed to manage the display.
#
my $scr = { };

d_start($scr, $nrows, $ncols);

d_disp_labels($scr);

# The color mapping is screwy.  On OSX the terminal type that works
# is 'xterm' so before running this program, in the bash wrapper
# we do
#   export TERM='xterm'
#
if ( 0 ) {
    # Debug the damn color settings.  
    d_stop();
    printf "\n";
    printf "ncolors %d npairs %d\n", $scr->{ncolors}, $scr->{npairs};
    printf "colors: %s\n", join(" ", d_colors( $scr ) );
    printf "\n";
    exit;
    }

# Every cell has an associated highlight that is adjusted during the
# animation.  The color pairs below are the highlights we use.

my $c_norm = d_hl( $scr, 'black' );
my $c_read = d_hl( $scr, 'green' );
my $c_write = d_hl( $scr, 'red' );
my $c_pc = d_hl( $scr, 'blue' );

# Initialize the memory values on the screen, and set c_norm
# as the defauly highlight.

foreach (my $i = 0; $i < $mem_max; $i++ ) {
    d_disp_addr($scr, $i, $mem->[$i], $c_norm);
    }

# $undo is a list of pairs [ $addr, $value ] that captures the address
# and value of every cell that was altered in processing the current
# instruction. This way we can revert the attributes back to normal
# when finished processing the instruction.

my $undo = [ ];

# Running state:
# SS - single step on s character
# NS - n steps
# C - continuous, .1 sec delay per fetch
# CD - continuous, 1 sec delay between updates

my $state = "SS";
my $nsteps = 0;

# Current step in the trace, should usually match $idx, but handled
# independently in case missing trace data.

my $step = 0;

my $prev_pc;
my $prev_inst_num;

RUN: for (my $step = 0; $step < $trace_step_max; $step++ ) {

    my ( $idx, $inst_num, $pc, $d, $addr, $value ) = @{$trace->[$step]};

    my $inst_num_changed = 
        ! defined $prev_inst_num || $prev_inst_num != $inst_num;

    # When we start a new instruction, undo the old display values
    if ( $inst_num_changed ) {
        printf DEBUG "starting new instruction, number %s\n", $inst_num;

        foreach my $u ( @$undo ) {
            d_disp_addr($scr, $u->[0], $u->[1], $c_norm);
            }
        
        $undo = [];
        $prev_inst_num = $inst_num;

        $state =~ /^C/ and delay(0.25);
        }

    # Top info area of screen contains:
    # INST# and instruction
    # IN
    # OUT
    # PC

    # Update the current instruction number - need to clear to eol to
    # write over any old chars left on line.

    addstr(0, 1, 
        sprintf("INST# %d: %s %s %s %s", $inst_num,
        $mem->[$pc], $mem->[$pc+1], $mem->[$pc+2], $mem->[$pc+3], 
        ) );
    clrtoeol;
    # addstr(1, 1, sprintf("IN") );
    # addstr(2, 1, sprintf("OUT") );

    # Turn off highlight of the cell pointed to by previous pc
    if ( defined $prev_pc && $pc != $prev_pc ) {
        d_disp_addr($scr, $prev_pc, $mem->[$prev_pc], $c_norm);
        }

    # Highlight the cell pointed to by current pc
    push @$undo, @{d_disp_addr($scr, $pc, $mem->[$pc], $c_pc)};

    # and the next 3 so instruction is clear
    $inst_num_changed and
    map {
        push @$undo, @{d_disp_addr($scr, $pc+$_, $mem->[$pc+$_], $c_pc)} 
        } (1, 2, 3);

    # Highlight cell accessed accoring to the operation
    # Also update memory to reflect change due to write

    my $hl;
    if ( $d eq "W" ) {
        $mem->[$addr] = $value;
        $hl = $c_write;
        }
    else {
        $hl = $c_read;
        }

    # Record the highlights we need to undo.
    push @$undo, @{d_disp_addr($scr, $addr, $value, $hl)};
    $prev_pc = $pc;

    0 and
    printf DEBUG "undo is now \n%s\n", join("\n",
        map { "  " . $_->[0] . ":" . $_->[1] } @$undo );

    if ( $state eq "NS" ) {
        if ( $inst_num_changed ) {
            $nsteps--;

            delay(0.25);

            $nsteps <= 0 and $state = "SS";
            }
        }
    
    elsif ( $state eq "C" ) {
        delay(0.1);
        $inst_num_changed and
            $nsteps--;

        $nsteps <= 0 and
            $state = "SS";

        }

    elsif ( $state eq "SS" ) {
        if ( $inst_num_changed ) {
            $nsteps = 0;
            }
        else {
            $nsteps++;
            }

        my $ch = getchar() // "";

        last RUN if $ch =~ /^q/i;

        if ( $ch =~ /c/i ) {
            $state = "C";
            $nsteps = 10;
            }

        elsif ( $ch =~ /n/i ) {
            # run for n instructions
            $nsteps = getstring() // "0";

            printf DEBUG "read nsteps as %s\n", $nsteps;

            my $ok = 
                Scalar::Util::looks_like_number($nsteps) && $nsteps > 0;

            if ( $ok ) {
                $state = "NS";
                }
            else {
                $state = "SS";
                $nsteps = 0;
                }
            }
        }
    else {
        # unknown state
        d_msg( $scr, sprintf("Unknown state '%s'", $state) );
        next RUN;
        }

    d_msg( $scr, $state . " " . $nsteps );

    } # end RUN

d_stop($scr);

#
#   Useful functions
#

sub delay
{
    # High resolution sleep that allows delay in microsecond resolution
    my ($sec) = @_;

    Time::HiRes::sleep($sec);
    }

#
#   Curses screen ADT
#
# $I is a hash with these fields:
#
#   ncolors => number of available colors
#   npairs  => number of defined colors pairs
#   color => { map of color name to color pair index }
#
#   nrows  => number of rows in memory cell area
#   ncols  => number of ncolsin memory cell area
#
#   col_width => number of chars in a cell column
#   col_offset => offset from left of first cell in a row.
#   row_offset => offset from top of first row in the cell matrix

sub d_start
{
    my ($I, $nrows, $ncols) = @_;

    initscr;
    initscr;
    cbreak;
    noecho;

    nonl;
    #  intrflush(stdscr, FALSE);
    #  keypad(stdscr, TRUE);


    die "Terminal has no colors" unless
        has_colors;

    start_color;
    use_default_colors;

    my $ncolors = $I->{ncolors} = COLORS;
    my $npairs = $I->{npairs} = COLOR_PAIRS;

    for (my $i=2; $i < $npairs; $i++ ) {
        init_pair($i, $i % $ncolors, -1);
        }

    # Color attribute pairs 
    # 2 green
    # 3 yelllow
    # 4 blue
    # 5 magenta
    # 6 cyan
    # 7 grey
    # 8 black
    # 9 red

    $I->{nrows} = $nrows;
    $I->{ncols} = $ncols;

    $I->{color} = {
        green => 2,
        yellow => 3,
        blue => 4,
        magenta => 5,
        cyan => 6,
        grey => 7,
        black => 8,
        red => 9,
        };

    # Cell array offsets
    $I->{col_width} = 5;
    $I->{col_offset} = 8;
    $I->{row_offset} = 7;
    }

sub d_colors
{
    my ($I) = @_;
    return sort keys %{$I->{color}};
    }

sub d_hl
{
    # Map a color name to a Curses color attribute pair
    my ($I, $name) = @_;

    my $pair = $I->{color}->{black};
    
    exists $I->{color}->{$name} and
        $pair = $I->{color}->{$name};

    return $pair;
    }

sub d_stop
{
    endwin;
    system("reset");
    printf "Should be back to normal\n";
    }

sub d_msg
{
    my ($I, $str) = @_;

    # update the msg bar to $str
    addstr(5, 1, sprintf("MSG: %s", $str) ); 
    clrtoeol;
    }

sub d_disp_addr
{
    # Update the display of memory cell $addr with value $value and
    # color attribute $hlight.

    my ($I, $addr, $value, $hlight) = @_;

    my $undo = [ ];

    my $col_width = $I->{col_width};
    my $col_offset = $I->{col_offset};
    my $row_offset = $I->{row_offset};

    # Handle PC, IN, OUT addresses in upper panel, skip other negative
    # addresses.

    return $undo if $addr < -4;

    if ( $addr == -1 ) {
        # IN
        my $dval = defined $value ?
            sprintf("IN: %d", $value) : sprintf("IN: ?");
        addstr(1, 1, $dval);
        clrtoeol;
        }
    elsif ( $addr == -2 ) {
        # OUT
        my $dval = defined $value ?
            sprintf("OUT: %d", $value) : sprintf("OUT: ?");
        addstr(2, 1, $dval);
        clrtoeol;
        }
    elsif ( $addr == -3 ) {
        # CIN
        my $dval = defined $value ?
            sprintf("CIN: %d '%s'", $value, chr($value)) : sprintf("CIN: ?");
        addstr(1, 1, $dval);
        clrtoeol;
        }
    elsif ( $addr == -4 ) {
        # COUT
        my $dval = defined $value ?
            sprintf("COUT: %d '%s'", $value, int_to_str($value)) : sprintf("COUT: ?");
        addstr(2, 1, $dval);
        clrtoeol;
        }
    else {
        # Also show PC in special place
        if ( $addr == 0 ) {
            my $dval = defined $value ?
                sprintf("PC: %d%10s", $value, " ") : sprintf("PC: ?%10s", " ");
            addstr(3, 1, $dval);
            }

        my $row = $row_offset + int ( $addr / $I->{ncols} );
        my $col = $col_offset + $col_width * ($addr % $I->{ncols});

        attron(COLOR_PAIR($hlight));
        my $dval = defined $value ?
            sprintf(" %4d", $value) : sprintf(" %4s", "?");
        
        addstr($row, $col, $dval);
        attroff(COLOR_PAIR($hlight));

        # Record for attribute undo
        push @$undo, [ $addr, $value ];
        }

    refresh;
    return $undo;
    }

sub int_to_str
{
    my ($x) = @_;
    my $s = "";
    while ( $x > 0 ) {
        $s = chr($x % 256) . $s;
        $x = int( $x / 256 );
        }
    return $s;
    }

sub d_disp_labels
{
    # Label the rows and cols of the cell matrix with address info

    my ($I) = @_;

    my $col_width = $I->{col_width};
    my $col_offset = $I->{col_offset};
    my $row_offset = $I->{row_offset};
    my $nrows = $I->{nrows};
    my $ncols = $I->{ncols};

    for (my $c = 0; $c < $ncols; $c++ ) {
        addstr($row_offset-1, $col_offset + $c * $col_width, 
            sprintf("%*d", $col_width, $c) );
        } 

    my $indent = " " x ($col_offset - 6);
    for (my $r = 0; $r < $nrows; $r++ ) {
        addstr($row_offset + $r, 1, sprintf("%s%4d:", $indent, $r) );
        } 

    refresh;
    }
